"""The types used in the `cg_feedback_helpers` package.
"""

import typing as t
from enum import Enum


class ExitStrategy(Enum):
    """Possible exit strategies used by :class:`cg_feedback_helpers.assertions.Asserter` in case of
    negative feedback.
    """
    #: Use `sys.exit` in case of negative feedback. The negative feedback will
    #: be printed to the configured output.
    SYS_EXIT = 0
    #: Use exceptions in case of negative feedback. The negative feedback will
    #: be used as the exception message.
    RAISE = 1


# This class allows us to have a type for `NO_FEEDBACK` that can be used in the
# general `Feedback` UnionType.
class NoFeedback:  # pylint: disable=too-few-public-methods
    """Sentinel used to inform the helpers to not print any feedback.
    """

    # Coverage is disabled as we do not really care about what the string
    # representation of this is.
    def __repr__(self) -> str:  #pragma: no cover
        return "NoFeedback"


NO_FEEDBACK = NoFeedback()
"""Constant that represents the "do not print any feedback" case.
"""

Feedback = t.Union[str, NoFeedback]
"""- None, empty string = use default feedback
- non-empty string = custom feedback
- NoFeedback = Do not print any feedback
"""


class FeedbackType(Enum):
    """Currently supported feedback types.
    """
    #: Feedback provided in case of assertions passing.
    POSITIVE = 'positive'
    #: Feedback provided in case of assertions failing.
    NEGATIVE = 'negative'


class FeedbackPiece(t.TypedDict):
    """A plaintext piece of feedback. Currently, the package
    does not support markdown feedback.
    """
    #: The content of the feedback.
    value: str
    #: The sentiment the feedback should have (neutral not supported).
    sentiment: t.Literal['positive', 'negative']


class FeedbackMessage(t.TypedDict):
    """The message to be sent for each piece of feedback
    """
    #: The message tag.
    tag: t.Literal['feedback']
    #: The list of feedback pieces
    contents: t.List[FeedbackPiece]


BASE_FILTER_OUT_GLOBALS = [
    "__name__",
    "__doc__",
    "__package__",
    "__loader__",
    "__spec__",
    "__annotations__",
    "__builtins__",
    "__file__",
    "__cached__",
    "__warningregistry__",
    "asserter",
    "helpers",
    "cg_feedback_helpers",
]
